"""Interface for running an exported NengoEdge model in TFLite format."""

import warnings
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import tensorflow as tf

from nengo_edge import config


class TFLiteRunner:
    """
    Run an exported TFLite model.

    Parameters
    ----------
    directory : Union[str, Path]
        Path to the directory containing the exported model files.
    """

    def __init__(self, directory: Union[str, Path]):
        self.directory = Path(directory)
        self.model_params, self.preprocessing = config.load_params(self.directory)

        # Build interpreters
        model_path = self.directory / "model.tflite"
        if not model_path.exists():
            raise FileNotFoundError(f"Could not find model file ({model_path})")
        self.model_interpreter = tf.lite.Interpreter(model_path=str(model_path))
        self.model_interpreter.allocate_tensors()

        feature_path = self.directory / "feature_extractor.tflite"
        if not feature_path.exists():
            warnings.warn(
                f"Could not find feature extractor model file ({feature_path}).\n"
                f"Using the tflite runner without a compiled feature extractor is "
                f"an experimental feature and may lead to poor model performance."
            )
            self.feature_interpreter = None
        else:
            self.feature_interpreter = tf.lite.Interpreter(model_path=str(feature_path))
            self.feature_interpreter.allocate_tensors()

        self.reset_state()

    def reset_state(self, batch_size: int = 1) -> None:
        """Reset the internal state of the model to initial conditions."""
        assert self.model_interpreter is not None
        self.model_state = [
            np.zeros([batch_size] + x["shape"].tolist()[1:], dtype=x["dtype"])
            for x in self.model_interpreter.get_input_details()[1:]
        ]

        if self.feature_interpreter is not None:
            self.feature_state = [
                np.zeros(
                    [
                        # initialize dynamic state sizes to 0 to avoid startup artifacts
                        batch_size if i == 0 else (0 if sig == -1 else val)
                        for i, (sig, val) in enumerate(
                            zip(x["shape_signature"], x["shape"])
                        )
                    ],
                    dtype=x["dtype"],
                )
                for x in self.feature_interpreter.get_input_details()[1:]
            ]

    def run(self, inputs: np.ndarray) -> np.ndarray:
        """
        Run the model on the given inputs.

        Parameters
        ----------
        inputs : np.ndarray
            Model input values (should have shape ``(batch_size, input_steps)``).

        Returns
        -------
        outputs : ``np.ndarray``
            Model output values (with shape ``(batch_size, output_d)`` if
            ``self.model_params['return_sequences']=False``
            else ``(batch_size, output_steps, output_d)``).
        """
        if self.feature_interpreter is not None:
            inputs, self.feature_state = self._run_feature_extractor(inputs)

        outputs, self.model_state = self._run_model(inputs)

        return outputs

    def _resize_inputs(
        self, interpreter: tf.lite.Interpreter, input_vals: List[np.ndarray]
    ) -> None:
        reallocate = False
        for s0, s1 in zip(input_vals, interpreter.get_input_details()):
            if s0.shape != tuple(s1["shape"]):
                # this will attempt to resize any dynamic dimensions (e.g. batch
                # size), and will raise an error if there is a mismatch on
                # static dimensions
                interpreter.resize_tensor_input(s1["index"], s0.shape, strict=True)
                reallocate = True
        if reallocate:
            interpreter.allocate_tensors()

    def _run_feature_extractor(
        self, inputs: np.ndarray
    ) -> Tuple[np.ndarray, List[np.ndarray]]:
        """Run the feature extractor on the given inputs."""
        assert self.feature_interpreter is not None

        self._resize_inputs(self.feature_interpreter, [inputs] + self.feature_state)

        for inp, val in zip(
            self.feature_interpreter.get_input_details(), [inputs] + self.feature_state
        ):
            self.feature_interpreter.set_tensor(inp["index"], val.astype(np.float32))

        self.feature_interpreter.invoke()

        outputs = self.feature_interpreter.get_tensor(
            self.feature_interpreter.get_output_details()[0]["index"]
        )
        feature_state = [
            self.feature_interpreter.get_tensor(out["index"])
            for out in self.feature_interpreter.get_output_details()[1:]
        ]

        return outputs, feature_state

    def _run_model(self, inputs: np.ndarray) -> Tuple[np.ndarray, List[np.ndarray]]:
        """Run the main model on the given inputs."""

        n_unroll = self.model_interpreter.get_input_details()[0]["shape"][1]
        n_steps = inputs.shape[1]
        if n_steps % n_unroll != 0:
            raise ValueError(
                f"Number of timesteps ({n_steps}) cannot be evenly divided by unroll "
                f"size ({n_unroll})"
            )

        self._resize_inputs(
            self.model_interpreter, [inputs[:, :n_unroll]] + self.model_state
        )
        input_details = self.model_interpreter.get_input_details()
        output_details = self.model_interpreter.get_output_details()

        input_tensors = [
            self.model_interpreter.tensor(details["index"]) for details in input_details
        ]
        output_tensors = [
            self.model_interpreter.tensor(details["index"])
            for details in output_details
        ]

        # quantize all the inputs ahead of time (for efficiency)
        inputs = self.quantize(inputs, input_details[0])

        state_input = self.model_state
        if self.model_params["return_sequences"]:
            output_sequences = []

        for step in range(0, n_steps, n_unroll):
            # set input for current block of unrolled steps
            input_tensors[0]()[:] = inputs[:, step : step + n_unroll]
            for inp, fp32_data, details in zip(
                input_tensors[1:], state_input, input_details[1:]
            ):
                inp()[:] = self.quantize(fp32_data, details)

            # run interpreter
            self.model_interpreter.invoke()

            # get state input for next block of steps
            state_input = [
                self.dequantize(out(), details)
                for out, details in zip(output_tensors[1:], output_details[1:])
            ]

            if self.model_params["return_sequences"]:
                # Append all outputs from the past `unroll` timesteps
                output_sequences.append(output_tensors[0]().copy())

        if self.model_params["return_sequences"]:
            # concatenate outputs from each unroll block
            # note: because the model was built with return_sequences=True, this
            # represents the output from all the timesteps
            outputs = np.concatenate(output_sequences, axis=1)
        else:
            # only use the output from last unroll block
            # note: because the model was built with return_sequences=False, the
            # output from the last unroll block is actually the output from the
            # last timestep
            outputs = output_tensors[0]().copy()

        # dequantize outputs (state_input was already dequantized in the loop above)
        outputs = self.dequantize(outputs, output_details[0])

        return outputs, state_input

    @staticmethod
    def quantize(val: np.ndarray, details: Dict[str, Any]) -> np.ndarray:
        """Quantize the given value based on quantization parameters."""
        n_scales = len(details["quantization_parameters"]["scales"])
        if n_scales == 0:
            # val is not quantized
            return np.array(val)

        assert n_scales == 1
        val = np.asarray(val)
        scaling_factor = details["quantization_parameters"]["scales"][0]
        zero_point = details["quantization_parameters"]["zero_points"][0]
        iinfo = np.iinfo(details["dtype"])
        q = np.round((val / scaling_factor) + zero_point)
        return np.clip(q, iinfo.min, iinfo.max).astype(details["dtype"])

    @staticmethod
    def dequantize(val: np.ndarray, details: Dict[str, Any]) -> np.ndarray:
        """Dequantize the given value based on quantization parameters."""
        n_scales = len(details["quantization_parameters"]["scales"])
        if n_scales == 0:
            # val is not quantized
            return np.array(val)

        assert n_scales == 1
        scaling_factor = details["quantization_parameters"]["scales"][0]
        zero_point = details["quantization_parameters"]["zero_points"][0]
        return (val.astype(np.float32) - zero_point) * scaling_factor



# pylint: disable=missing-docstring

import struct
from pathlib import Path
from typing import Any, List, Optional

import numpy as np
import pytest
import serial

from nengo_edge.micro_runner import DiscoRunner, MicroRunner, NordicRunner


class MockSerial:
    def __init__(self, read_data: List[bytes], write_data: List[bytes]):
        self.read_data = read_data
        self.write_data = write_data
        self.read_count = 0
        self.write_count = 0
        self.timeout = 60

    def read(self, n_bytes: int) -> bytes:
        data = self.read_data[self.read_count]
        self.read_count += 1
        return data

    def write(self, data: bytes) -> None:
        assert data == self.write_data[self.write_count], (
            f"({self.write_count}) "
            f"Expected: {self.write_data[self.write_count]!r} "
            f"Actual: {data!r}"
        )
        self.write_count += 1

    def reset_input_buffer(self) -> None:
        pass

    def reset_output_buffer(self) -> None:
        pass

    def __enter__(self) -> "MockSerial":
        return self

    def __exit__(self, *_: Any) -> None:
        pass

    def close(self) -> None:
        pass


def mock_runner(
    tmp_path: Path,
    n_unroll: int,
    monkeypatch: pytest.MonkeyPatch,
    type: str,
) -> MicroRunner:
    device_drive = tmp_path / "micro_device_drive"
    device_drive.mkdir(parents=True, exist_ok=True)

    binary_path = tmp_path / (
        "nengoedge_project.bin" if type == "disco" else "nengoedge_project.hex"
    )
    binary_path.touch()

    def mock_send_data_burst(
        self: MicroRunner,
        serial_port: serial.Serial,
        user_data: np.ndarray,
        state_data: Optional[List[np.ndarray]],
    ) -> None:
        pass

    def mock_receive_data_burst(
        self: MicroRunner,
        serial_port: serial.Serial,
        size: int,
    ) -> np.ndarray:
        if self.model_params["return_sequences"]:
            rxd_data = np.ones((n_unroll * (size // n_unroll)), dtype="float32")
        else:
            rxd_data = np.ones((size), dtype="float32")

        return rxd_data

    monkeypatch.setattr(serial, "Serial", lambda *_, **__: MockSerial([], []))
    monkeypatch.setattr(MicroRunner, "send_data_burst", mock_send_data_burst)
    monkeypatch.setattr(MicroRunner, "receive_data_burst", mock_receive_data_burst)

    runner = (DiscoRunner if type == "disco" else NordicRunner)(
        directory=tmp_path,
        serial_path=Path("/dummy_serial_path"),
        device_path=device_drive,
    )

    return runner


def test_send_data_burst() -> None:
    # device fails to receive ack
    serial_port = MockSerial(
        # device sends request+success, then request+failure, then request+success
        read_data=[b"R", b"S", b"R", b"F", b"R", b"S"],
        # host should send ack, then data (first without state then with state)
        write_data=[
            b"N",
            struct.pack("<f", 0),
            b"T",
            struct.pack("<ff", 1, 2),
            b"T",
            struct.pack("<ff", 1, 2),
        ],
    )

    MicroRunner.send_data_burst(serial_port, np.asarray([0]), None)
    MicroRunner.send_data_burst(serial_port, np.asarray([1]), [np.asarray([2])])
    assert serial_port.read_count == len(serial_port.read_data)
    assert serial_port.write_count == len(serial_port.write_data)

    # device request is lost
    serial_port = MockSerial(
        # timeout on request read, then request+success
        read_data=[b"", b"R", b"S"],
        # host should send failure, then ack+data
        write_data=[b"F", b"N", struct.pack("<f", 0)],
    )
    MicroRunner.send_data_burst(serial_port, np.asarray([0]), None)
    assert serial_port.read_count == len(serial_port.read_data)
    assert serial_port.write_count == len(serial_port.write_data)

    # timeout on send
    serial_port = MockSerial([], [])
    serial_port.timeout = 0
    with pytest.raises(TimeoutError, match="trying to send data to device"):
        MicroRunner.send_data_burst(serial_port, np.asarray([0]), None)


def test_read_data_burst() -> None:
    # ack lost
    serial_port = MockSerial(
        # missing ack, then ack+data
        read_data=[b"", b"T", struct.pack("<f", 1)],
        # host should send req, then failure, then another req, then success
        write_data=[b"R", b"F", b"R", b"S"],
    )
    data = MicroRunner.receive_data_burst(serial_port, 1)
    assert np.array_equal(data, [1])
    assert serial_port.read_count == len(serial_port.read_data)
    assert serial_port.write_count == len(serial_port.write_data)

    # timeout on read
    serial_port = MockSerial([], [])
    serial_port.timeout = 0
    with pytest.raises(TimeoutError, match="trying to receive data from device"):
        MicroRunner.receive_data_burst(serial_port, 1)


def test_runner(
    param_dir: Path,
    rng: np.random.RandomState,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    batch_size = 3

    runner = mock_runner(
        tmp_path=param_dir, n_unroll=1, monkeypatch=monkeypatch, type="disco"
    )
    inputs = rng.uniform(-0.5, 0.5, size=(batch_size, 16000))
    with runner:
        # check that batched and non-batched models produce the
        # same output (state is being properly reset between batch items)
        out = []
        for i in range(batch_size):
            out.append(runner.run(inputs[i : i + 1]))
        out0 = np.concatenate(out, axis=0)

    with runner:
        out1 = runner.run(inputs)

    assert out0.shape == out1.shape == (batch_size, 49, 10)

    # increased tolerance due to float32 variance
    assert np.allclose(out0, out1, atol=1e-5), np.max(abs(out0 - out1))


def test_runner_errors(
    param_dir: Path,
    monkeypatch: pytest.MonkeyPatch,
    rng: np.random.RandomState,
) -> None:
    batch_size = 3

    runner = mock_runner(
        tmp_path=param_dir, n_unroll=1, monkeypatch=monkeypatch, type="nordic"
    )

    inputs = rng.uniform(-0.5, 0.5, size=(batch_size, 10))

    # check sample rate too low
    runner.preprocessing["sample_rate"] = 10
    with runner:
        with pytest.raises(ValueError, match="sample_rate >> 100"):
            runner.run(inputs)

    runner.preprocessing["sample_rate"] = 16000

    # check n_steps % n_unroll != 0
    runner.model_params["input_shape"][0] = 3

    inputs = rng.uniform(-0.5, 0.5, size=(16000,))
    with runner:
        with pytest.raises(ValueError, match="Inputs must have ndim=2"):
            runner.run(inputs)

        inputs = np.reshape(inputs, (1, 16000))
        with pytest.raises(ValueError, match="cannot be evenly divided by unroll"):
            runner.run(inputs)



"""On device runner for the Coral dev board."""

import argparse
import json
import platform
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Union

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    from nengo_edge.device_modules.np_mfcc import LogMelFeatureExtractor


class CoralDeviceRunner:  # pragma: no cover (runs on-device)
    """
    Run a compiled model on the Coral board.

    Parameters
    ----------
    directory : Path
        Directory containing the edgetpu TFLite model file that will be executed.
    return_sequences: bool
        Flag to return all outputs
    """

    def __init__(
        self,
        directory: Union[str, Path],
        return_sequences: bool,
    ):
        self.directory = Path(directory)
        self.return_sequences = return_sequences

        # Build interpreter
        # pylint: disable=import-outside-toplevel
        import tflite_runtime.interpreter as tflite

        model_files = list(self.directory.glob("*_edgetpu.tflite"))
        if len(model_files) == 0:
            raise FileNotFoundError(f"No model files in {self.directory}")
        elif len(model_files) > 1:
            raise ValueError(f"Found multiple model files {model_files}")

        self.interpreter = tflite.Interpreter(
            model_path=str(model_files[0]),
            experimental_delegates=[
                tflite.load_delegate(
                    {
                        "Linux": "libedgetpu.so.1",
                        "Darwin": "libedgetpu.1.dylib",
                        "Windows": "edgetpu.dll",
                    }[platform.system()],
                )
            ],
        )
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()

        self.interpreter.allocate_tensors()

        self.n_unroll = self.input_details[0]["shape"][1]

    def run(self, inputs: np.ndarray) -> np.ndarray:
        """
        Run the model on the given inputs.

        Parameters
        ----------
        inputs : np.ndarray
            Model input values (should have shape ``(batch_size, n_steps, input_d)``.

        Returns
        -------
            Model output values (with shape ``(batch_size, output_d)`` if
            ``self.return_sequences=False``  else
            ``(batch_size, output_steps, output_d)``).
        """

        assert inputs.shape[0] == 1
        n_steps = inputs.shape[1]
        assert n_steps % self.n_unroll == 0

        input_tensors = [
            self.interpreter.tensor(details["index"]) for details in self.input_details
        ]
        output_tensors = [
            self.interpreter.tensor(details["index"]) for details in self.output_details
        ]

        # quantize all the inputs ahead of time (for efficiency)
        inputs = self.quantize(inputs, self.input_details[0])

        state_input = [
            np.zeros(x["shape"], dtype=x["dtype"]) for x in self.input_details[1:]
        ]
        output_sequences = []
        for step in range(0, n_steps, self.n_unroll):
            # set input for current block of unrolled steps
            input_tensors[0]()[:] = inputs[:, step : step + self.n_unroll]
            for inp, fp32_data, details in zip(
                input_tensors[1:], state_input, self.input_details[1:]
            ):
                inp()[:] = self.quantize(fp32_data, details)

            # run interpreter
            self.interpreter.invoke()

            # get state input for next block of steps
            state_input = [
                self.dequantize(out(), details)
                for out, details in zip(output_tensors[1:], self.output_details[1:])
            ]

            if self.return_sequences:
                # Append all outputs from the past `unroll` timesteps
                output_sequences.append(output_tensors[0]().copy())

        outputs = (
            # concatenate outputs from each unroll block
            np.concatenate(output_sequences, axis=1)
            if self.return_sequences
            # only use the output from last unroll block
            else output_tensors[0]().copy()
        )

        # dequantize outputs
        outputs = self.dequantize(outputs, self.output_details[0])

        return outputs

    @staticmethod
    def quantize(val: np.ndarray, details: Dict[str, np.ndarray]) -> np.ndarray:
        """Quantize the given value based on quantization parameters."""

        val = np.asarray(val)
        assert len(details["quantization_parameters"]["scales"]) == 1
        assert len(details["quantization_parameters"]["zero_points"]) == 1
        scaling_factor = details["quantization_parameters"]["scales"][0]
        zero_point = details["quantization_parameters"]["zero_points"][0]
        iinfo = np.iinfo(details["dtype"])
        q = np.round((val / scaling_factor) + zero_point)
        return np.clip(q, iinfo.min, iinfo.max).astype(details["dtype"])

    @staticmethod
    def dequantize(val: np.ndarray, details: Dict[str, np.ndarray]) -> np.ndarray:
        """Dequantize the given value based on quantization parameters."""

        assert len(details["quantization_parameters"]["scales"]) == 1
        assert len(details["quantization_parameters"]["zero_points"]) == 1
        scaling_factor = details["quantization_parameters"]["scales"][0]
        zero_point = details["quantization_parameters"]["zero_points"][0]
        return (val.astype(np.float32) - zero_point) * scaling_factor


def _build_feature_extractor(
    args: argparse.Namespace,
) -> "LogMelFeatureExtractor":  # pragma: no cover (runs on-device)
    # pylint: disable=import-outside-toplevel

    from np_mfcc import LogMelFeatureExtractor  # type: ignore

    with open(Path(args.directory) / "parameters.json", "r", encoding="utf-8") as f:
        parameters = json.load(f)

    options = parameters["preprocessing"]
    feature_extractor = LogMelFeatureExtractor(**options)

    return feature_extractor


def _run_model(
    feature_extractor: "LogMelFeatureExtractor",
    runner: CoralDeviceRunner,
    inputs: np.ndarray,
) -> np.ndarray:
    inputs = np.concatenate(
        [feature_extractor(inputs[i : i + 1]) for i in range(inputs.shape[0])],
        axis=0,
    )

    # coral board only supports batch size of 1, so we have to manually iterate
    # over batches
    out_samples = []
    for i in range(inputs.shape[0]):
        out_samples.append(runner.run(inputs=inputs[i : i + 1]))
    out = np.concatenate(out_samples, axis=0)

    return out


def cli() -> None:
    """Command line interface for running on device."""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--directory",
        required=True,
        help=(
            "Path to build artifacts on coral board"
            " (including the edgetpu model and parameters.json files)"
        ),
    )
    parser.add_argument(
        "--return-sequences",
        action="store_true",
        help="Return all outputs generated by the model",
    )
    args = parser.parse_args()

    # set up feature extractor
    feature_extractor = _build_feature_extractor(args)

    # build model/interpreter
    runner = CoralDeviceRunner(
        directory=Path(args.directory), return_sequences=args.return_sequences
    )

    # load inputs from file
    with np.load(Path(args.directory) / "inputs.npz", allow_pickle=False) as data:
        inputs = data["inputs"]

    # run model
    out = _run_model(feature_extractor, runner, inputs)

    # save outputs to file
    np.save(Path(args.directory) / "outputs.npy", out, allow_pickle=False)


if __name__ == "__main__":  # pragma: no cover (runs on-device)
    cli()



"""Interface for loading parameter file associated with a compiled model and runner."""

import json
from pathlib import Path
from typing import Any, Dict, Tuple


def load_params(directory: Path) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Load parameters from file."""

    param_path = directory / "parameters.json"
    if not param_path.exists():
        raise FileNotFoundError(f"Could not find parameter file ({param_path})")
    with open(param_path, "r", encoding="utf-8") as f:
        params = json.load(f)
        model_params = params["model"]
        preprocessing = params["preprocessing"]

    return model_params, preprocessing



# Originally based on
# https://github.com/google-coral/project-keyword-spotter/blob/master/model.py
# https://github.com/google-coral/project-keyword-spotter/blob/master/mel_features.py
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Extract MFCC speech features from audio data.

This is designed to work with minimal python dependencies (numpy only) so that it
can run on on-device CPUs.
"""


from typing import Union

import numpy as np


def frame(data: np.ndarray, window_length: int, hop_length: int) -> np.ndarray:
    """
    Convert array into a sequence of successive possibly overlapping frames.

    An n-dimensional array of shape (batch_size, num_samples, ...) is converted into an
    (n+1)-D array of shape (batch_size, num_frames, window_length, ...), where each
    frame starts hop_length points after the preceding one.

    This is accomplished using stride_tricks, so the original data is not
    copied.  However, there is no zero-padding, so any incomplete frames at the
    end are not included.

    Parameters
    ----------
    data : np.array
        Data to be split into frames, with shape ``(batch_size, num_samples, ...)``.
    window_length : int
        Number of samples in each frame.
    hop_length : int
        Advance (in samples) between each frame.

    Returns
    -------
    frames : np.ndarray
        (N+1)-D np.array with as many rows as there are complete frames that can be
        extracted.
    """
    num_samples = data.shape[1]
    num_frames = 1 + int(np.floor((num_samples - window_length) / hop_length))
    shape = (data.shape[0], num_frames, window_length) + data.shape[2:]
    strides = data.strides[:1] + (data.strides[1] * hop_length,) + data.strides[1:]
    return np.lib.stride_tricks.as_strided(data, shape=shape, strides=strides)


def periodic_hann(window_length: int) -> np.ndarray:
    """
    Calculate a "periodic" Hann window.

    The classic Hann window is defined as a raised cosine that starts and
    ends on zero, and where every value appears twice, except the middle
    point for an odd-length window.  Matlab calls this a "symmetric" window
    and np.hanning() returns it.  However, for Fourier analysis, this
    actually represents just over one cycle of a period N-1 cosine, and
    thus is not compactly expressed on a length-N Fourier basis.  Instead,
    it's better to use a raised cosine that ends just before the final
    zero value - i.e. a complete cycle of a period-N cosine.  Matlab
    calls this a "periodic" window. This routine calculates it.

    Parameters
    ----------
    window_length : int
        The number of points in the returned window.

    Returns
    -------
    window : np.ndarray
        A 1D array containing the periodic hann window.
    """
    return 0.5 - (0.5 * np.cos(2 * np.pi / window_length * np.arange(window_length)))


def stft_magnitude(
    signal: np.ndarray, fft_length: int, hop_length: int, window_length: int
) -> np.ndarray:
    """
    Calculate the short-time Fourier transform magnitude.

    Parameters
    ----------
    signal : np.ndarray
        2D np.array of the batched input time-domain signal.
    fft_length : int
        Size of the FFT to apply.
    hop_length : int
        Advance (in samples) between each frame passed to FFT.
    window_length : int
        Length of each block of samples to pass to FFT.

    Returns
    -------
    spectrogram : np.ndarray
        3D np.array where the first axis is the batch, the second axis is the number
        of frames, and the third axis contains the magnitudes of the fft_length/2+1
        unique values of the FFT for the corresponding frame of input samples.
    """
    frames = frame(signal, window_length, hop_length)
    # Apply frame window to each frame. We use a periodic Hann (cosine of period
    # window_length) instead of the symmetric Hann of np.hanning (period
    # window_length-1).
    window = periodic_hann(window_length)
    windowed_frames = frames * window[None, :]
    return np.abs(np.fft.rfft(windowed_frames, int(fft_length)))


def hertz_to_mel(
    frequencies_hertz: "Union[float, np.ndarray]",
) -> "Union[float, np.ndarray]":
    """
    Convert frequencies to mel scale using HTK formula.

    Parameters
    ----------
    frequencies_hertz : Union[float, np.ndarray]
        Scalar or np.array of frequencies in hertz.

    Returns
    -------
    mel_values : Union[float, np.ndarray]
        Object of same size as ``frequencies_hertz`` containing corresponding values
        on the mel scale.
    """

    # Mel spectrum constants and functions.
    _MEL_BREAK_FREQUENCY_HERTZ = 700.0
    _MEL_HIGH_FREQUENCY_Q = 1127.0

    return _MEL_HIGH_FREQUENCY_Q * np.log(
        1.0 + (frequencies_hertz / _MEL_BREAK_FREQUENCY_HERTZ)
    )


def spectrogram_to_mel_matrix(
    num_mel_bins: int = 20,
    num_spectrogram_bins: int = 129,
    audio_sample_rate: int = 8000,
    lower_edge_hertz: float = 125.0,
    upper_edge_hertz: float = 3800.0,
) -> np.ndarray:
    """
    Return a matrix that can post-multiply spectrogram rows to make mel.

    Returns a np.array matrix A that can be used to post-multiply a matrix S of
    spectrogram values (STFT magnitudes) arranged as frames x bins to generate a
    "mel spectrogram" M of frames x num_mel_bins.  M = S A.

    The classic HTK algorithm exploits the complementarity of adjacent mel bands
    to multiply each FFT bin by only one mel weight, then add it, with positive
    and negative signs, to the two adjacent mel bands to which that bin
    contributes.  Here, by expressing this operation as a matrix multiply, we go
    from num_fft multiplies per frame (plus around 2*num_fft adds) to around
    num_fft^2 multiplies and adds.  However, because these are all presumably
    accomplished in a single call to np.dot(), it's not clear which approach is
    faster in Python.  The matrix multiplication has the attraction of being more
    general and flexible, and much easier to read.

    Parameters
    ----------
    num_mel_bins : int
        How many bands in the resulting mel spectrum.  This is
        the number of columns in the output matrix.
    num_spectrogram_bins : int
        How many bins there are in the source spectrogram
        data, which is understood to be fft_size/2 + 1, i.e. the spectrogram
        only contains the nonredundant FFT bins.
    audio_sample_rate : int
        Samples per second of the audio at the input to the
        spectrogram. We need this to figure out the actual frequencies for
        each spectrogram bin, which dictates how they are mapped into mel.
    lower_edge_hertz : float
        Lower bound on the frequencies to be included in the mel
        spectrum.  This corresponds to the lower edge of the lowest triangular
        band.
    upper_edge_hertz : float
        The desired top edge of the highest frequency band.

    Returns
    -------
    mel_matrix : np.ndarray
        An array with shape (num_spectrogram_bins, num_mel_bins) that can post-multiply
        the spectrogram to compute mel values.

    Raises
    ------
    ValueError
        If frequency edges are incorrectly ordered or out of range.
    """
    nyquist_hertz = audio_sample_rate / 2.0
    if lower_edge_hertz < 0.0:
        raise ValueError(f"lower_edge_hertz {lower_edge_hertz:.1f} must be >= 0")
    if lower_edge_hertz >= upper_edge_hertz:
        raise ValueError(
            f"lower_edge_hertz {lower_edge_hertz:.1f} >= upper_edge_hertz "
            f"{upper_edge_hertz:.1f}"
        )
    if upper_edge_hertz > nyquist_hertz:
        raise ValueError(
            f"upper_edge_hertz {upper_edge_hertz:.1f} is greater than Nyquist "
            f"{nyquist_hertz:.1f}"
        )
    spectrogram_bins_hertz = np.linspace(0.0, nyquist_hertz, num_spectrogram_bins)
    spectrogram_bins_mel = hertz_to_mel(spectrogram_bins_hertz)
    # The i'th mel band (starting from i=1) has center frequency
    # band_edges_mel[i], lower edge band_edges_mel[i-1], and higher edge
    # band_edges_mel[i+1].  Thus, we need num_mel_bins + 2 values in
    # the band_edges_mel arrays.
    band_edges_mel = np.linspace(
        hertz_to_mel(lower_edge_hertz), hertz_to_mel(upper_edge_hertz), num_mel_bins + 2
    )
    # Matrix to post-multiply feature arrays whose rows are num_spectrogram_bins
    # of spectrogram values.
    mel_weights_matrix = np.empty((num_spectrogram_bins, num_mel_bins))
    for i in range(num_mel_bins):
        lower_edge_mel, center_mel, upper_edge_mel = band_edges_mel[i : i + 3]
        # Calculate lower and upper slopes for every spectrogram bin.
        # Line segments are linear in the *mel* domain, not hertz.
        lower_slope = (spectrogram_bins_mel - lower_edge_mel) / (
            center_mel - lower_edge_mel
        )
        upper_slope = (upper_edge_mel - spectrogram_bins_mel) / (
            upper_edge_mel - center_mel
        )
        # .. then intersect them with each other and zero.
        mel_weights_matrix[:, i] = np.maximum(0.0, np.minimum(lower_slope, upper_slope))
    # HTK excludes the spectrogram DC bin; make sure it always gets a zero
    # coefficient.
    mel_weights_matrix[0, :] = 0.0
    return mel_weights_matrix


def dct_matrix(input_features: int, output_features: int) -> np.ndarray:
    """Returns a matrix that can post-multiply mel features to compute DCT features."""

    matrix = 2.0 * np.cos(
        np.pi
        * np.outer(
            np.arange(input_features) * 2.0 + 1.0,
            np.arange(output_features),
        )
        / (2.0 * input_features)
    )
    # DCT normalization
    matrix *= 1.0 / np.sqrt(2.0 * input_features)

    return matrix


class LogMelFeatureExtractor:
    """
    Compute log mel spectrogram slices from audio data.

    Parameters
    ----------
    window_size_ms : float
        Length of each spectrogram time slice (in milliseconds).
    window_stride_ms : float
        Length of time (in milliseconds) between spectrogram slices.
    mel_num_bins : int
        The number of Mel bins.
    dct_num_features : int
        The number of DCT features.
    sample_rate : int
        Sample rate (in Hz) of the audio data.
    mel_lower_edge_hertz : float
        Lower bound on the frequencies to be included in the mel spectrum.
    mel_upper_edge_hertz : float
        Upper bound on the frequencies to be included in the mel spectrum.
    log_epsilon : float
        Small value added to avoid NaNs in log calculation.
    """

    def __init__(
        self,
        window_size_ms: float,
        window_stride_ms: float,
        mel_num_bins: int,
        dct_num_features: int,
        sample_rate: int,
        mel_lower_edge_hertz: float,
        mel_upper_edge_hertz: float,
        log_epsilon: float,
    ):
        self._norm_factor = 3
        self.window_size_samples = int(round(sample_rate * window_size_ms / 1000.0))
        self.window_stride_samples = int(round(sample_rate * window_stride_ms / 1000.0))
        self.mel_num_bins = mel_num_bins
        self.dct_num_features = dct_num_features
        self.fft_length = 2 ** int(
            np.ceil(np.log(self.window_size_samples) / np.log(2.0))
        )
        self.log_epsilon = log_epsilon

        self.mel_matrix = (
            spectrogram_to_mel_matrix(
                num_mel_bins=mel_num_bins,
                num_spectrogram_bins=self.fft_length // 2 + 1,
                audio_sample_rate=sample_rate,
                lower_edge_hertz=mel_lower_edge_hertz,
                upper_edge_hertz=mel_upper_edge_hertz,
            ),
        )

        if dct_num_features:
            self.dct_matrix = dct_matrix(mel_num_bins, dct_num_features)

    def __call__(self, data: np.ndarray) -> np.ndarray:
        """
        Computes spectrograms on a batch of input data (not streaming).

        Parameters
        ----------
        data : np.ndarray
            Input data with shape ``(batch_size, audio_samples)``.

        Returns
        -------
        mfccs : np.ndarray
            Output mel spectrograms with shape
            ``(batch_size, frames, dct_num_features)`` (or ``(..., mel_num_features)``
            if ``dct_num_features==0``).
        """

        # compute windowing + fft
        features = stft_magnitude(
            data,
            fft_length=self.fft_length,
            hop_length=self.window_stride_samples,
            window_length=self.window_size_samples,
        )

        # compute mel features
        features = np.matmul(features, self.mel_matrix)

        # compute log mel features
        features = np.log(np.maximum(features, self.log_epsilon))

        # compute dct
        if self.dct_num_features:
            features = np.matmul(features, self.dct_matrix)

        return features



"""Interface for running an exported NengoEdge model in SavedModel format."""

from pathlib import Path
from typing import Dict, Union

import numpy as np
import tensorflow as tf


class SavedModelRunner:
    """Run a model exported in TensorFlow's SavedModel format."""

    def __init__(self, directory: Union[str, Path]):
        self.directory = Path(directory)

        self.model = tf.saved_model.load(str(self.directory)).signatures[
            "serving_default"
        ]

        # The saved model takes inputs and returns outputs organized by name. But in
        # general we don't know what those names will be, because they can depend on
        # what other models have been loaded in the TensorFlow graph. However, when
        # creating the models in NengoEdge we ensure that the inputs/outputs will be
        # ordered alphabetically, so we can use that to recover the correct
        # input/output order here.
        self.input_names = sorted(self.model.structured_input_signature[1])
        self.output_names = sorted(self.model.structured_outputs)

        self.reset_state()

    def reset_state(self) -> None:
        """Reset the internal state of the model to initial conditions."""

        self.state: Dict[str, tf.Tensor] = {}

    def run(self, inputs: np.ndarray) -> np.ndarray:
        """
        Run the model on the given inputs.

        Parameters
        ----------
        inputs : np.ndarray
            Model input values (should have shape ``(batch_size, input_steps)``).

        Returns
        -------
        outputs : ``np.ndarray``
            Model output values (with shape ``(batch_size, output_d)`` if
            the model was built to return only the final time step,
            else ``(batch_size, output_steps, output_d)``).
        """

        inputs = tf.cast(inputs, "float32")

        kwargs = {}
        for name, sig in self.model.structured_input_signature[1].items():
            if name == self.input_names[0]:
                kwargs[name] = inputs
            else:
                if name not in self.state:
                    self.state[name] = tf.zeros(
                        [inputs.shape[0]]
                        + [0 if s is None else s for s in sig.shape[1:]]
                    )
                kwargs[name] = self.state[name]

        outputs = self.model(**kwargs)

        # Update saved state
        for input_name, output_name in zip(self.input_names[1:], self.output_names[1:]):
            self.state[input_name] = outputs[output_name]

        return outputs[self.output_names[0]]



# pylint: disable=missing-docstring

import runpy

from nengo_edge import version


def test_version_string() -> None:
    version_string = runpy.run_path(version.__file__)["version"]

    assert ("dev" in version_string) == (version.dev is not None)
    assert version_string == version.version



# pylint: disable=missing-docstring

import numpy as np
import pytest
import tensorflow as tf
from nengo_edge_models.audio import SpeechFeatures
from nengo_edge_models.audio.base_data import MFCCOptions

from nengo_edge.device_modules import np_mfcc


def test_frame(rng: np.random.RandomState) -> None:
    x = rng.uniform(0, 1, size=(32, 16000))

    y0 = tf.signal.frame(x, 400, 160)

    y1 = np_mfcc.frame(x, 400, 160)

    assert np.allclose(y0, y1)


def test_hann(rng: np.random.RandomState) -> None:
    y0 = tf.signal.hann_window(400, periodic=True, dtype=tf.float64)
    y1 = np_mfcc.periodic_hann(400)

    assert np.allclose(y0, y1)


def test_mel_matrix() -> None:
    y0 = tf.signal.linear_to_mel_weight_matrix(dtype=tf.float64)
    y1 = np_mfcc.spectrogram_to_mel_matrix()

    assert np.allclose(y0, y1)


def test_spectrogram(rng: np.random.RandomState) -> None:
    x = rng.uniform(0, 1, size=(32, 16000))

    options = MFCCOptions(
        window_size_ms=400,
        window_stride_ms=160,
        sample_rate=1000,
        dct_num_features=0,
    )
    layer = SpeechFeatures(options)
    y0 = layer.spectrogram(x)

    y1 = np_mfcc.stft_magnitude(
        x,
        fft_length=2 ** int(np.ceil(np.log(400) / np.log(2.0))),
        hop_length=160,
        window_length=400,
    )

    assert np.allclose(y0, y1), np.max(abs(y0 - y1))


def test_mel_spectrogram(rng: np.random.RandomState) -> None:
    x = rng.uniform(0, 1, size=(32, 16000))

    options = MFCCOptions()

    layer = SpeechFeatures(options, dtype="float64")
    y0 = layer(x)

    feature_extractor = np_mfcc.LogMelFeatureExtractor(
        window_size_ms=options.window_size_ms,
        window_stride_ms=options.window_stride_ms,
        mel_num_bins=options.mel_num_bins,
        dct_num_features=options.dct_num_features,
        sample_rate=options.sample_rate,
        mel_lower_edge_hertz=options.mel_lower_edge_hertz,
        mel_upper_edge_hertz=options.mel_upper_edge_hertz,
        log_epsilon=options.log_epsilon,
    )
    y1 = feature_extractor(x)

    assert np.allclose(y0, y1), np.max(abs(y0 - y1))


def test_dct(rng: np.random.RandomState) -> None:
    x = rng.uniform(0, 1, size=(32, 49, 40))
    x = tf.cast(x, tf.float32)
    layer = SpeechFeatures(MFCCOptions(dct_num_features=20))
    layer.build(x.shape)
    y0 = layer.dct(x)

    dct_matrix = np_mfcc.dct_matrix(40, 20)
    y1 = np.matmul(x, dct_matrix)

    assert np.allclose(y0, y1, atol=1e-5), np.max(abs(y0 - y1))


def test_mfcc_features(rng: np.random.RandomState) -> None:
    options = MFCCOptions()
    x = rng.uniform(0, 2, size=(32, 16000))

    feature_layer = SpeechFeatures(options, dtype=tf.float64)
    y0 = feature_layer(x)

    feature_extractor = np_mfcc.LogMelFeatureExtractor(
        window_size_ms=options.window_size_ms,
        window_stride_ms=options.window_stride_ms,
        mel_num_bins=options.mel_num_bins,
        dct_num_features=options.dct_num_features,
        sample_rate=options.sample_rate,
        mel_lower_edge_hertz=options.mel_lower_edge_hertz,
        mel_upper_edge_hertz=options.mel_upper_edge_hertz,
        log_epsilon=options.log_epsilon,
    )
    y1 = feature_extractor(x)

    assert np.allclose(y0, y1), np.max(abs(y0 - y1))


def test_errors() -> None:
    with pytest.raises(ValueError, match="must be >= 0"):
        np_mfcc.spectrogram_to_mel_matrix(lower_edge_hertz=-1)

    with pytest.raises(ValueError, match=">= upper_edge_hertz"):
        np_mfcc.spectrogram_to_mel_matrix(lower_edge_hertz=1, upper_edge_hertz=1)

    with pytest.raises(ValueError, match="greater than Nyquist"):
        np_mfcc.spectrogram_to_mel_matrix(upper_edge_hertz=4001, audio_sample_rate=8000)



# pylint: disable=missing-docstring

from pathlib import Path

import pytest

from nengo_edge import config


def test_runner_errors(tmp_path: Path) -> None:
    with pytest.raises(FileNotFoundError, match="Could not find parameter file"):
        config.load_params(tmp_path)



import json

import pytest


@pytest.fixture
def param_dir(tmp_path):
    params = {
        "preprocessing": {
            "window_size_ms": 40,
            "window_stride_ms": 20,
            "mel_num_bins": 40,
            "dct_num_features": 20,
            "sample_rate": 16000,
            "mel_lower_edge_hertz": 20,
            "mel_upper_edge_hertz": 7000,
            "log_epsilon": 1e-12,
        },
        "model": {
            "input_shape": [1, 20],
            "output_shape": [1, 10],
            "state_shapes": [10, 10],
            "return_sequences": True,
        },
    }

    with open(tmp_path / "parameters.json", "w", encoding="utf-8") as f:
        json.dump(params, f)

    return tmp_path



"""Tools for working with NengoEdge."""

from nengo_edge.device_modules.coral_device import CoralDeviceRunner
from nengo_edge.micro_runner import DiscoRunner, NordicRunner
from nengo_edge.network_runner import CoralRunner
from nengo_edge.saved_model_runner import SavedModelRunner
from nengo_edge.tflite_runner import TFLiteRunner
from nengo_edge.version import copyright as __copyright__
from nengo_edge.version import version as __version__



# pylint: disable=missing-docstring

from pathlib import Path

import numpy as np
import pytest
import tensorflow as tf
from nengo_edge_hw import coral
from nengo_edge_models.kws.tests.test_models import _test_lmu

from nengo_edge.tflite_runner import TFLiteRunner


@pytest.mark.parametrize("return_sequences", (True, False))
def test_runner_streaming(
    return_sequences: bool,
    tmp_path: Path,
    rng: np.random.RandomState,
    monkeypatch: pytest.MonkeyPatch,
    seed: int,
) -> None:
    batch_size = 3
    tf.keras.utils.set_random_seed(seed)

    # export a test model
    monkeypatch.setattr(coral.host.Interface, "io_dtype", None)
    host = coral.host.Interface(
        _test_lmu(),
        build_dir=tmp_path,
        use_device=False,
        return_sequences=return_sequences,
    )

    host.export_model(tmp_path / "exported")

    inputs = rng.uniform(-0.5, 0.5, size=(batch_size, 16000))

    runner = TFLiteRunner(tmp_path / "exported")

    # check that batched and non-batched models produce the same output (state is being
    # properly reset between batch items)
    out = []
    for i in range(batch_size):
        out.append(runner.run(inputs[i : i + 1]))
        runner.reset_state(batch_size=1)
    out0 = np.concatenate(out, axis=0)

    runner.reset_state(batch_size=batch_size)
    out1 = runner.run(inputs)

    assert (
        out0.shape
        == out1.shape
        == ((batch_size, 49, 10) if return_sequences else (batch_size, 10))
    )

    # increased tolerance due to float32 variance
    assert np.allclose(out0, out1, atol=1e-5), np.max(abs(out0 - out1))

    # check that running sequentially produces the same output as running all at once
    runner.reset_state(batch_size=batch_size)
    if return_sequences:
        out0 = np.concatenate(
            [runner.run(inputs[:, :4000]), runner.run(inputs[:, 4000:])],
            axis=1,
        )
    else:
        runner.run(inputs[:, :4000])
        out0 = runner.run(inputs[:, 4000:])
    runner.reset_state(batch_size=batch_size)
    out1 = runner.run(inputs)
    assert (
        out0.shape
        == out1.shape
        == ((batch_size, 49, 10) if return_sequences else (batch_size, 10))
    )
    assert np.allclose(out0, out1, atol=5e-5), np.max(abs(out0 - out1))


def test_runner_errors(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    model_desc = _test_lmu()
    model_desc.n_unroll = 2
    monkeypatch.setattr(coral.host.Interface, "io_dtype", None)
    host = coral.host.Interface(
        model_desc, build_dir=tmp_path, use_device=False, mode="model-only"
    )
    host.export_model(tmp_path)

    runner = TFLiteRunner(tmp_path)

    with pytest.raises(ValueError, match="evenly divided by unroll"):
        runner.run(np.zeros((1, 3, 10)))

    Path(tmp_path / "model.tflite").unlink()

    with pytest.raises(FileNotFoundError, match="Could not find model file"):
        TFLiteRunner(tmp_path)


def test_runner_quantized(tmp_path: Path, rng: np.random.RandomState) -> None:
    model_desc = _test_lmu()
    host = coral.host.Interface(
        model_desc,
        build_dir=tmp_path,
        use_device=False,
        representative_data=rng.uniform(-1, 1, size=(32, 16000)),
    )
    host.export_model(tmp_path)

    runner = TFLiteRunner(tmp_path)

    x = rng.uniform(-1, 1, size=(1, 16000))
    y0 = runner.run(x)
    y1 = host.run(x)

    assert np.allclose(y0, y1, atol=5e-5), np.max(abs(y0 - y1))



# pylint: disable=missing-docstring

import argparse
import json
import sys
from pathlib import Path
from typing import Union

import numpy as np
import pytest

from nengo_edge import network_runner
from nengo_edge.device_modules import coral_device, np_mfcc
from nengo_edge.tests.test_network_runner import MockRunner


class MockDeviceRunner(coral_device.CoralDeviceRunner):
    def __init__(  # pylint: disable=super-init-not-called
        self, directory: Union[str, Path], return_sequences: bool
    ) -> None:
        self.directory = Path(directory)
        self.return_sequences = return_sequences

    def run(self, inputs: np.ndarray) -> np.ndarray:
        return inputs


def mock_feature_extractor(
    args: argparse.Namespace,
) -> np_mfcc.LogMelFeatureExtractor:
    # note: this is the same as coral_device._build_feature_extractor,
    # with this import removed
    # from np_mfcc import LogMelFeatureExtractor

    with open(Path(args.directory) / "parameters.json", "r", encoding="utf-8") as f:
        parameters = json.load(f)

    options = parameters["preprocessing"]
    feature_extractor = np_mfcc.LogMelFeatureExtractor(**options)
    return feature_extractor


def test_cli(
    monkeypatch: pytest.MonkeyPatch,
    param_dir: Path,
    rng: np.random.RandomState,
) -> None:
    batch_size = 3
    inputs = rng.uniform(-0.5, 0.5, size=(batch_size, 16000)).astype("float32")

    monkeypatch.setattr(network_runner, "CoralRunner", MockRunner)
    monkeypatch.setattr(coral_device, "CoralDeviceRunner", MockDeviceRunner)
    monkeypatch.setattr(
        coral_device, "_build_feature_extractor", mock_feature_extractor
    )

    binary_path = param_dir / "model_edgetpu.tflite"
    binary_path.touch()

    runner = MockRunner(directory=param_dir, username="user", hostname="hostname")

    # get output from cli
    assert isinstance(runner.remote_dir, Path)
    runner.prepare_device_runner()
    runner.send_inputs(inputs)

    monkeypatch.setattr(
        sys,
        "argv",
        f"coral_runner.py --directory {runner.remote_dir}".split(),
    )
    coral_device.cli()

    cli_output = np.load(runner.remote_dir / "outputs.npy")

    options = runner.preprocessing
    feature_extractor = np_mfcc.LogMelFeatureExtractor(**options)
    features = feature_extractor(inputs)

    assert np.all(cli_output == features)



"""
Runner for running an exported NengoEdge model in TFLite format on microcontroller
devices.

NengoEdge supports interfacing with STM32F746 Discovery and nRF52840 Dev boards
via this runner.
"""
import shutil
import time
import timeit
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import serial
from numpy.typing import NDArray
from rich import progress

from nengo_edge import config


class MicroRunner:
    """
    Run an exported model from nengo edge that runs on micro devices.

    Parameters
    ----------
    directory : Union[str, Path]
        Path to the directory containing the nengo edge export artifacts.
    serial_path : Union[str, Path]
        Path to drive where device serial port is mounted
    device_path : Union[str, Path]
        Path to drive where device is mounted
    binary_name : str
        Name for the binary file used by this device.
    """

    # control bytes for serial communication
    SUCCESS = b"S"
    REQ = b"R"
    ACK_NOSTATE = b"N"
    ACK_STATE = b"T"
    FAIL = b"F"

    def __init__(
        self,
        directory: Union[str, Path],
        serial_path: Union[str, Path],
        device_path: Union[str, Path],
        binary_name: str,
    ):
        self.directory = Path(directory)
        self.serial_path = Path(serial_path)
        self.device_path = Path(device_path)
        self.binary_name = binary_name
        self.model_params, self.preprocessing = config.load_params(self.directory)
        self.uses_states = len(self.model_params["state_shapes"]) > 0

        # the amount of data that will be output by the model
        self.recv_size = self.model_params["output_shape"][-1]

        # Flash the device binary
        self.flash_binary()

    def __enter__(self) -> "MicroRunner":
        self.serial_port = serial.Serial(
            str(self.serial_path), baudrate=115200, timeout=30
        )
        time.sleep(0.1)  # Add a slight delay to make sure serial port is ready
        return self

    def __exit__(self, *_: Any) -> None:
        self.serial_port.close()

    def flash_binary(self) -> None:
        """Copy the binary to the device."""

        shutil.copyfile(
            self.directory / self.binary_name, self.device_path / self.binary_name
        )

        # Wait for 10s for device to be programmed with copied binary to avoid
        # race condition.
        for _ in progress.track(range(10), description="Flashing binary to device"):
            time.sleep(1)

    def prime_inputs(self, inputs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Prepare inputs to match expected audio buffer in the model binary."""

        # We need to prime the audio buffer with window - stride samples so we
        # can properly extract a full window of MFCCs given stride new samples
        n_priming = (
            (
                self.preprocessing["window_size_ms"]
                - self.preprocessing["window_stride_ms"]
            )
            * self.preprocessing["sample_rate"]
            / 1000
        )

        if not np.isclose(n_priming, int(n_priming)):
            raise ValueError(
                f"Sample rate from parameters ({self.preprocessing['sample_rate']})"
                f" is too small. Use sample_rate >> 100."
            )

        n_priming = min(int(n_priming), inputs.shape[1])
        primer_inputs = inputs[:, :n_priming]
        inputs = inputs[:, n_priming:]

        inputs = inputs.reshape(
            (
                inputs.shape[0],
                -1,
                int(
                    self.preprocessing["sample_rate"]
                    * 1e-3
                    * self.preprocessing["window_stride_ms"]
                ),
            )
        )

        return inputs, primer_inputs

    def prepare_serial(self, inputs: np.ndarray) -> None:
        """Send required runtime parameters before any inputs."""

        n_steps = np.shape(inputs)[1]
        n_unroll = self.model_params["input_shape"][0]

        # Make sure data can be pieced into unroll sized chunks
        if n_steps % n_unroll != 0:
            raise ValueError(
                f"Number of timesteps ({n_steps}) cannot be evenly divided by unroll "
                f"size ({n_unroll})"
            )

        if self.model_params["return_sequences"]:
            self.recv_size *= n_unroll

        # Binary expects number of loops it will execute at runtime (sent over UART)
        self.n_hw_laps = len(inputs) * n_steps // n_unroll
        self.send_data_burst(
            serial_port=self.serial_port,
            user_data=np.asarray([float(self.n_hw_laps)]),
            state_data=None,
        )

    def run(self, inputs: np.ndarray) -> np.ndarray:
        """
        Run model inference on a batch of inputs.

        Parameters
        ----------
        inputs : np.ndarray
            Model input values (must have shape ``(batch_size, input_steps)``)

        Returns
        -------
        outputs : ``np.ndarray``
            Model output values (with shape ``(batch_size, output_d)`` if
            ``self.model_params['return_sequences']=False``
            else ``(batch_size, output_steps, output_d)``).
        """
        if len(inputs.shape) != 2:
            raise ValueError("Inputs must have ndim=2 (batch_size, input_steps).")

        inputs, primer_inputs = self.prime_inputs(inputs)
        self.prepare_serial(inputs)

        outputs = self._run_model(inputs, primer_inputs=primer_inputs)

        self.serial_port.reset_input_buffer()

        return outputs

    def _run_model(
        self, inputs: np.ndarray, primer_inputs: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Run the main model logic on the given inputs."""

        n_steps = np.shape(inputs)[1]
        n_unroll = self.model_params["input_shape"][0]

        outputs = []
        with progress.Progress() as bar:
            task = bar.add_task("Running inference on device", total=self.n_hw_laps)
            for index_batch, user_data in enumerate(inputs):
                # Reshape this input to be in unroll sized chunks.
                user_data = np.reshape(user_data, (n_steps // n_unroll, -1))
                # Create list of initial state data for this input (or None if no
                # model)
                state_data = (
                    [np.zeros(shape) for shape in self.model_params["state_shapes"]]
                    if self.uses_states
                    else None
                )

                # Send in audio primer inputs
                if primer_inputs is not None:
                    # We send initial state with the first slice below for
                    # "model-only", so neglect state here and send state below for
                    # consistency.
                    self.send_data_burst(
                        serial_port=self.serial_port,
                        user_data=primer_inputs[index_batch].reshape(-1),
                        state_data=None,
                    )

                # Send each slice over and receive output for that slice
                output_sequences = []
                for index_slice, input_slice in enumerate(user_data):
                    self.send_data_burst(
                        serial_port=self.serial_port,
                        user_data=input_slice,
                        state_data=None
                        if index_slice > 0 or not self.uses_states
                        else state_data,
                    )
                    rxd_data = self.receive_data_burst(
                        serial_port=self.serial_port, size=self.recv_size
                    )

                    if self.model_params["return_sequences"]:
                        # Append all outputs from the past `unroll` timesteps
                        # Need to reshape since they come as block
                        # n_unroll*n_outputs
                        output_sequences.append(
                            rxd_data.reshape(
                                n_unroll,
                                self.recv_size // n_unroll,
                            )
                        )

                    bar.update(task, advance=1)

                # Store final output or all outputs if return_sequences
                outputs.append(
                    # concatenate outputs from each unroll block
                    # note: because the model was built with return_sequences=True,
                    # this represents the output from all the timesteps
                    np.concatenate(output_sequences, axis=0)
                    if self.model_params["return_sequences"]
                    # only use the output from last unroll block
                    # note: because the model was built with return_sequences=False,
                    # the output from the last unroll block is actually the output
                    # from the last timestep
                    else rxd_data
                )

        return np.asarray(outputs)

    @staticmethod
    def send_data_burst(
        serial_port: serial.Serial,
        user_data: np.ndarray,
        state_data: Optional[List[np.ndarray]],
    ) -> None:
        """
        Send the given burst of data over UART to remote device.

        Parameters
        ----------
        serial_port : serial.Serial
            The serial object used to send data
        user_data : np.ndarray
            Model input values (should have shape ``(n_unroll * input_d)``
        state_data : Optional[List[np.ndarray]]
            Initial state values. Each LMU layer has two state values, for the
            hidden (with shape ``(units)``) and memory (with shape
            ``(memory_d * order)``) components. A multi-layer model
            should repeat these in the same order as the layers, i.e.
            ``[memory_state_0, hidden_state_0, memory_state_1, hidden_state_1, ...]``.
        """

        # Only send state if necessary
        flattened_data = user_data
        if state_data is not None:
            flattened_data = np.append(
                flattened_data, np.concatenate(state_data, axis=None)
            )

        timeout = timeit.default_timer() + serial_port.timeout * 5
        reply = b""
        while reply != MicroRunner.SUCCESS and timeit.default_timer() < timeout:
            data = serial_port.read(1)

            if data == MicroRunner.REQ:
                serial_port.write(
                    MicroRunner.ACK_NOSTATE
                    if state_data is None
                    else MicroRunner.ACK_STATE
                )

                serial_port.write(flattened_data.astype(np.float32).tobytes())
                reply = serial_port.read(1)
            else:
                # this will cause the device to re-send the send request
                serial_port.write(MicroRunner.FAIL)

        if reply != MicroRunner.SUCCESS:
            raise TimeoutError("Timed out trying to send data to device")

    @staticmethod
    def receive_data_burst(serial_port: serial.Serial, size: int) -> NDArray:
        """
        Receive burst of data over UART from remote device.

        Parameters
        ----------
        serial_port : serial.Serial
            The serial object used to send data
        size: int
            The count of FP32 numbers to receive from remote device.
        """
        timeout = timeit.default_timer() + serial_port.timeout * 5
        success = False
        while not success and timeit.default_timer() < timeout:
            serial_port.reset_input_buffer()
            serial_port.write(MicroRunner.REQ)

            data = serial_port.read(1)
            if data == MicroRunner.ACK_STATE:
                data = serial_port.read(size * 4)
                if len(data) == size * 4:
                    data = np.frombuffer(data, dtype=np.float32)
                    success = True

            # send response
            serial_port.write(MicroRunner.SUCCESS if success else MicroRunner.FAIL)

        if not success:
            raise TimeoutError("Timed out trying to receive data from device")

        return data


class DiscoRunner(MicroRunner):
    """
    Run a model exported from NengoEdge on the Disco board.

    See `MicroRunner` for parameter descriptions.
    """

    def __init__(
        self,
        directory: Union[str, Path],
        serial_path: Union[str, Path],
        device_path: Union[str, Path],
    ):
        super().__init__(
            directory=directory,
            serial_path=serial_path,
            device_path=device_path,
            binary_name="nengoedge_project.bin",
        )


class NordicRunner(MicroRunner):
    """
    Run a model exported from NengoEdge on the Nordic board.

    See `MicroRunner` for parameter descriptions.
    """

    def __init__(
        self,
        directory: Union[str, Path],
        serial_path: Union[str, Path],
        device_path: Union[str, Path],
    ):
        super().__init__(
            directory=directory,
            serial_path=serial_path,
            device_path=device_path,
            binary_name="nengoedge_project.hex",
        )



"""
Interface for running an exported NengoEdge model on network accessible devices.

Nengo-edge supports running on the Coral dev board via this runner.
"""

import subprocess
from pathlib import Path
from typing import List, Union

import numpy as np

from nengo_edge import config
from nengo_edge.device_modules import coral_device, np_mfcc


class NetworkRunner:
    """
    Run an exported model from nengo edge that runs on remote network devices.

    Parameters
    ----------
    directory : Union[str, Path]
        Path to the directory containing the nengo edge export artifacts.
    device_runner : Union[str, Path]
        Path to the file that will run on the device.
    username : str
        Username on the remote device
    hostname : str
        Hostname of the remote device
    """

    def __init__(
        self,
        directory: Union[str, Path],
        device_runner: Union[str, Path],
        username: str,
        hostname: str,
    ):
        self.directory = Path(directory)
        self.device_runner = Path(device_runner)
        self.model_params, self.preprocessing = config.load_params(self.directory)
        self.return_sequences = self.model_params["return_sequences"]

        self.username = username
        self.hostname = hostname
        self.address = f"{self.username}@{self.hostname}"
        self.check_connection()

        self.remote_dir = Path("/tmp/nengo-edge-runner")

        self.reset()

    def check_connection(self) -> None:
        """
        Run ssh to confirm connection to remote device.

        Throws runtime error if connection fails.
        """
        try:
            subprocess.run(f"ssh {self.address} echo ok".split(), check=True)
        except Exception as e:
            raise RuntimeError(
                f"Cannot connect to address {self.address}: {e!r}"
            ) from e

    def _scp(
        self, files_to_copy: List[Path]
    ) -> None:  # pragma: no cover (needs device)
        """One liner to send specified files to remote device location."""
        cmd = (
            f"scp {' '.join(str(p) for p in files_to_copy)} "
            f"{self.address}:{self.remote_dir}"
        )

        subprocess.run(cmd.split(), check=True)

    def reset(self) -> None:
        """Reset the state of the network runner."""
        self.prepared = False

    def send_inputs(self, inputs: np.ndarray) -> None:
        """Saves inputs to file and sends to device."""

        # save inputs to file
        filepath = self.directory / "inputs.npz"
        np.savez_compressed(filepath, inputs=inputs)

        # copy to device
        self._scp([filepath])

    def prepare_device_runner(self) -> None:  # pragma: no cover (needs device)
        """Send required runtime parameters/modules before any inputs."""

        subprocess.run(
            f"ssh {self.address} mkdir -p {self.remote_dir}".split(), check=True
        )

        # copy files to remote
        self._scp(
            [
                self.device_runner,
                self.directory / "model_edgetpu.tflite",
                self.directory / "parameters.json",
                Path(np_mfcc.__file__),
            ]
        )
        self.prepared = True

    def _run_model(
        self,
        inputs: np.ndarray,
    ) -> np.ndarray:  # pragma: no cover (needs device)
        """Run the main model logic on the given inputs."""
        self.send_inputs(inputs)
        subprocess.run(
            f"ssh {self.address} python3 {self.remote_dir / self.device_runner.name} "
            f"--directory {self.remote_dir} "
            f"{'--return-sequences' if self.return_sequences else ''}".split(),
            check=True,
        )

        # copy outputs back to host
        subprocess.run(
            f"scp {self.address}:{self.remote_dir / 'outputs.npy'} "
            f"{self.directory}".split(),
            check=True,
        )

        outputs = np.load(self.directory / "outputs.npy")
        return outputs

    def run(self, inputs: np.ndarray) -> np.ndarray:  # pragma: no cover (needs device)
        """
        Run model inference on a batch of inputs.

        Parameters
        ----------
        inputs : np.ndarray
            Model input values (must have shape ``(batch_size, input_steps)``)

        Returns
        -------
        outputs : ``np.ndarray``
            Model output values (with shape ``(batch_size, output_d)`` if
            ``self.model_params['return_sequences']=False``
            else ``(batch_size, output_steps, output_d)``).
        """
        if not self.prepared:
            self.prepare_device_runner()

        outputs = self._run_model(inputs)
        return outputs


class CoralRunner(NetworkRunner):
    """
    Run a model exported from NengoEdge on the Coral board.

    See `NetworkRunner` for parameter descriptions.
    """

    def __init__(
        self,
        directory: Union[str, Path],
        username: str,
        hostname: str,
    ):
        super().__init__(
            directory=directory,
            username=username,
            hostname=hostname,
            device_runner=coral_device.__file__,
        )



# pylint: disable=missing-docstring

from pathlib import Path
from typing import Literal

import numpy as np
import pytest
import tensorflow as tf
from nengo_edge_hw import gpu
from nengo_edge_models.kws.models import lmu_small

from nengo_edge.saved_model_runner import SavedModelRunner


@pytest.mark.parametrize("mode", ("model-only", "feature-only", "full"))
def test_runner(
    mode: Literal["model-only", "feature-only", "full"],
    rng: np.random.RandomState,
    seed: int,
    tmp_path: Path,
) -> None:
    tf.keras.utils.set_random_seed(seed)

    interface = gpu.host.Interface(lmu_small(), mode=mode, build_dir=tmp_path)

    inputs = rng.uniform(
        -1, 1, size=(32,) + ((49, 20) if mode == "model-only" else (16000,))
    ).astype("float32")

    output0 = interface.run(inputs)

    interface.export_model(tmp_path)

    assert interface.binary_path is not None
    runner = SavedModelRunner(tmp_path)

    output1 = runner.run(inputs)

    assert np.allclose(output0, output1), np.max(np.abs(output0 - output1))


def test_runner_streaming(
    rng: np.random.RandomState, seed: int, tmp_path: Path
) -> None:
    tf.keras.utils.set_random_seed(seed)

    interface = gpu.host.Interface(lmu_small(), mode="full", build_dir=tmp_path)

    # 3200 timesteps is equivalent to 200ms at 16 kHz
    inputs = rng.uniform(-1, 1, size=(32, 3200)).astype("float32")
    output0 = interface.run(inputs)

    interface.export_model(tmp_path / "streaming", streaming=True)
    runner = SavedModelRunner(tmp_path / "streaming")

    # check that running in parts produces the same output
    stream_chunk_size = inputs.shape[1] // 4
    for i in range(4):
        output1 = runner.run(
            inputs[:, i * stream_chunk_size : (i + 1) * stream_chunk_size]
        )

    assert np.allclose(output0, output1, atol=1e-4), np.max(np.abs(output0 - output1))

    # check that resetting state works
    runner.reset_state()
    for i in range(4):
        output2 = runner.run(
            inputs[:, i * stream_chunk_size : (i + 1) * stream_chunk_size]
        )

    assert np.allclose(output0, output2, atol=1e-4), np.max(np.abs(output0 - output2))

    # test zero padding
    runner.reset_state()
    pad_output = runner.run(inputs[:, :10])
    pad_output_gt = (
        interface.run(
            np.concatenate(
                [
                    inputs[:, :10],
                    np.zeros((32, interface.audio_options.window_size_samples - 10)),
                ],
                axis=1,
            )
        ),
    )
    assert np.allclose(pad_output, pad_output_gt, atol=2e-6), np.max(
        np.abs(pad_output - pad_output_gt)
    )
    pad_output_step = runner.run(inputs[:, -10:])
    pad_output_step_gt = interface.run(
        np.concatenate(
            [
                inputs[:, :10],
                np.zeros((32, interface.audio_options.window_size_samples - 10)),
                inputs[:, -10:],
                np.zeros((32, interface.audio_options.window_stride_samples - 10)),
            ],
            axis=1,
        )
    )
    assert np.allclose(pad_output_step, pad_output_step_gt, atol=2e-6), np.max(
        np.abs(pad_output_step - pad_output_step_gt)
    )



# pylint: disable=missing-docstring

import subprocess
from pathlib import Path
from typing import List

import numpy as np
import pytest

from nengo_edge import CoralRunner
from nengo_edge.device_modules import np_mfcc


class MockRunner(CoralRunner):
    def _scp(self, files_to_copy: List[Path]) -> None:
        subprocess.run(
            f"cp -r {' '.join(str(p) for p in files_to_copy)} "
            f"{self.remote_dir}".split(),
            check=True,
        )

    def check_connection(self) -> None:
        pass

    def prepare_device_runner(self) -> None:
        # this is the same as the super implementation, but with the ssh mkdir replaced
        # with a local mkdir
        # subprocess.run(
        #     f"ssh {self.address} mkdir -p {self.remote_dir}".split(), check=True
        # )
        self.remote_dir.mkdir(exist_ok=True, parents=True)

        self._scp(
            [
                self.device_runner,
                self.directory / "model_edgetpu.tflite",
                self.directory / "parameters.json",
                Path(np_mfcc.__file__),
            ]
        )
        self.prepared = True


def test_coral_runner(
    monkeypatch: pytest.MonkeyPatch,
    param_dir: Path,
    rng: np.random.RandomState,
) -> None:
    binary_path = param_dir / "model_edgetpu.tflite"
    binary_path.touch()

    with pytest.raises(RuntimeError, match="Cannot connect to address"):
        CoralRunner(directory=param_dir, username="user", hostname="host")

    net_runner = MockRunner(directory=param_dir, username="user", hostname="host")

    batch_size = 3
    inputs = rng.uniform(-0.5, 0.5, size=(batch_size, 16000)).astype("float32")

    # get output from cli
    assert isinstance(net_runner.remote_dir, Path)
    net_runner.prepare_device_runner()

    assert (net_runner.remote_dir / "parameters.json").exists()
    assert (net_runner.remote_dir / "np_mfcc.py").exists()

    net_runner.send_inputs(inputs)
    assert (net_runner.remote_dir / "inputs.npz").exists()



# Automatically generated by nengo-bones, do not edit this file directly

# pylint: disable=consider-using-f-string,bad-string-format-type

"""
NengoEdge Tools version information.

We use calendar versioning (see https://calver.org/) and conform to PEP440 (see
https://www.python.org/dev/peps/pep-0440/). '.dev0' will be added to the version
unless the code base represents a release version. Release versions are git
tagged with the version.
"""

version_info = (23, 7, 30)

name = "nengo-edge"
dev = None

# use old string formatting, so that this can still run in Python <= 3.5
# (since this file is parsed in setup.py, before python_requires is applied)
version = ".".join(str(v) for v in version_info)
if dev is not None:
    version += ".dev%d" % dev  # pragma: no cover

copyright = "Copyright (c) 2022-2023 Applied Brain Research"
